from .neat_dynamic_resnet import NEATTResNet
from .neat_focus import NEATFocus
from .neat_focus_microscope import NEATFocusPredict
from .neat_lstm import NEATLRNet
from .neat_microscope import NEATPredict
from .neat_static_resnet import NEATResNet
from .neat_vollnet import NEATVollNet
from .neat_densevollnet import NEATDenseVollNet
from .nets import Concat

# imports


__all__ = (
    "NEATLRNet",
    "NEATTResNet",
    "NEATVollNet",
    "NEATDenseVollNet",
    "NEATFocus",
    "NEATFocusPredict",
    "NEATPredict",
    "NEATResNet",
    "Concat",
)


#!/usr/bin/env python3
"""
Created on Wed Aug  4 14:50:47 2021

@author: vkapoor
"""
import csv
import glob
import json
import os
from pathlib import Path


import numpy as np
import pandas as pd
import napari
from qtpy.QtCore import Qt
from oneat.NEATUtils.oneat_animation._qt import OneatVolumeWidget, OneatWidget
from scipy import spatial
from tifffile import imread

from oneat.NEATUtils.oneat_animation.OneatVisualization import MidSlices


class NEATViz:
    def __init__(
        self,
        imagedir: str,
        csvdir: str,
        savedir: str,
        categories_json: dict,
        heatmapimagedir: str = None,
        segimagedir: str = None,
        heatname: str = "_Heat",
        eventname: str = "_Event",
        fileextension: str = "*tif",
        blur_radius: int = 5,
        start_project_mid: int = None,
        end_project_mid: int = None,
        headless: bool = False,
        volume: bool = True,
        batch: bool = False,
        event_threshold: float = 0.999,
        nms_space: int = 10,
        nms_time: int = 3,
    ):

        self.imagedir = imagedir
        self.heatmapimagedir = heatmapimagedir
        self.segimagedir = segimagedir
        self.savedir = savedir
        self.volume = volume
        self.csvdir = csvdir
        self.heatname = heatname
        self.eventname = eventname
        self.headless = headless
        self.batch = batch
        self.event_threshold = event_threshold
        self.nms_space = nms_space
        self.nms_time = nms_time
        self.categories_json = categories_json
        self.start_project_mid = start_project_mid
        self.end_project_mid = end_project_mid
        self.fileextension = fileextension
        self.blur_radius = blur_radius

        Path(self.savedir).mkdir(exist_ok=True)
        Path(self.csvdir).mkdir(exist_ok=True)
        

        self.time = 0
        self.key_categories = self.load_json()
        if not self.headless:
            
            self.viewer = napari.Viewer() 
        if not self.headless and not self.volume:
            self.showNapari()
        if self.headless and not self.volume:
            self.donotshowNapari()
        if self.volume and not self.headless:
            self.showVolumeNapari()
        if self.volume and self.headless:
            self.donotshowVolumeNapari()    

    def load_json(self):
        with open(self.categories_json) as f:
            return json.load(f)

    def donotshowVolumeNapari(self):
        
        headlessvolumecall(
                self.key_categories,
                self.event_threshold,
                self.nms_space,
                self.nms_time,
                self.csvdir,
                self.savedir,
            )
        
    def donotshowNapari(self):

        headlesscall(
                
                self.key_categories,
                self.event_threshold,
                self.nms_space,
                self.nms_time,
                self.csvdir,
                self.savedir,
            )

    def showNapari(self):

        self.oneat_widget = OneatWidget(
            self.viewer,
            self.csvdir,
            self.savedir,
            "Name",
            self.key_categories,
            segimagedir=self.segimagedir,
            heatimagedir=self.heatmapimagedir,
            heatname=self.heatname,
            start_project_mid=self.start_project_mid,
            end_project_mid=self.end_project_mid,
        )
        Raw_path = os.path.join(self.imagedir, self.fileextension)
        X = glob.glob(Raw_path)
        Imageids = []
        self.oneat_widget.frameWidget.imageidbox.addItem("Select Image")
        self.oneat_widget.frameWidget.eventidbox.addItem("Select Event")
        for imagename in X:
            Imageids.append(imagename)

        for i in range(0, len(Imageids)):
            self.oneat_widget.frameWidget.imageidbox.addItem(str(Imageids[i]))

        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                self.oneat_widget.frameWidget.eventidbox.addItem(event_name)

        dock_widget = self.viewer.window.add_dock_widget(
            self.oneat_widget, area="right"
        )
        self.viewer.window._qt_window.resizeDocks(
            [dock_widget], [200], Qt.Horizontal
        )

        napari.run()

    def showVolumeNapari(self):

        self.oneat_widget = OneatVolumeWidget(
            self.viewer,
            self.csvdir,
            self.savedir,
            "Name",
            self.key_categories,
            segimagedir=self.segimagedir,
            heatimagedir=self.heatmapimagedir,
            heatname=self.heatname,
        )
        Raw_path = os.path.join(self.imagedir, self.fileextension)
        X = glob.glob(Raw_path)
        Imageids = []
        self.oneat_widget.frameWidget.imageidbox.addItem("Select Image")
        self.oneat_widget.frameWidget.eventidbox.addItem("Select Event")
        for imagename in X:
            Imageids.append(imagename)

        for i in range(0, len(Imageids)):
            self.oneat_widget.frameWidget.imageidbox.addItem(str(Imageids[i]))

        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                self.oneat_widget.frameWidget.eventidbox.addItem(event_name)

        dock_widget = self.viewer.window.add_dock_widget(
            self.oneat_widget, area="right"
        )
        self.viewer.window._qt_window.resizeDocks(
            [dock_widget], [200], Qt.Horizontal
        )

        napari.run()


def cluster_points(
    event_locations_dict, event_locations_size_dict, nms_space, nms_time
):

    for (k, v) in event_locations_dict.items():
        currenttime = k
        event_locations = v

        tree = spatial.cKDTree(event_locations)
        for i in range(1, nms_time):

            forwardtime = currenttime + i
            if int(forwardtime) in event_locations_dict.keys():
                forward_event_locations = event_locations_dict[
                    int(forwardtime)
                ]
                for location in forward_event_locations:
                    if (
                        int(forwardtime),
                        int(location[0]),
                        int(location[1]),
                    ) in event_locations_size_dict:
                        forwardsize, forwardscore = event_locations_size_dict[
                            int(forwardtime),
                            int(location[0]),
                            int(location[1]),
                        ]
                        distance, nearest_location = tree.query(location)
                        nearest_location = int(
                            event_locations[nearest_location][0]
                        ), int(event_locations[nearest_location][1])

                        if distance <= nms_space:
                            if (
                                int(currenttime),
                                int(nearest_location[0]),
                                int(nearest_location[1]),
                            ) in event_locations_size_dict:
                                (
                                    currentsize,
                                    currentscore,
                                ) = event_locations_size_dict[
                                    int(currenttime),
                                    int(nearest_location[0]),
                                    int(nearest_location[1]),
                                ]
                                if currentsize >= forwardsize:
                                    event_locations_size_dict.pop(
                                        (
                                            int(forwardtime),
                                            int(location[0]),
                                            int(location[1]),
                                        )
                                    )

                                if currentsize < forwardsize:
                                    event_locations_size_dict.pop(
                                        (
                                            int(currenttime),
                                            int(nearest_location[0]),
                                            int(nearest_location[1]),
                                        )
                                    )
    return event_locations_size_dict

def cluster_spheres(event_locations_dict, event_locations_size_dict, nms_space, nms_time):

        print("before", len(event_locations_size_dict))

        for (k, v) in event_locations_dict.items():
            currenttime = k
            event_locations = v

            if len(event_locations) > 0:
                tree = spatial.cKDTree(event_locations)
                forwardtime = currenttime + 1
                if int(forwardtime) in event_locations_dict.keys():
                    forward_event_locations = event_locations_dict[
                        int(forwardtime)
                    ]
                    for location in forward_event_locations:
                        if (
                            int(forwardtime),
                            int(location[0]),
                            int(location[1]),
                            int(location[2]),
                        ) in event_locations_size_dict:
                            (
                                forwardsize,
                                forwardscore,
                                forwardconfidence,
                            ) = event_locations_size_dict[
                                int(forwardtime),
                                int(location[0]),
                                int(location[1]),
                                int(location[2]),
                            ]
                            distance, nearest_location = tree.query(location)
                            nearest_location = (
                                int(event_locations[nearest_location][0]),
                                int(event_locations[nearest_location][1]),
                                int(event_locations[nearest_location][2]),
                            )

                            if distance <= nms_space:
                                if (
                                    int(currenttime),
                                    int(nearest_location[0]),
                                    int(nearest_location[1]),
                                    int(nearest_location[2]),
                                ) in event_locations_size_dict:
                                    (
                                        currentsize,
                                        currentscore,
                                        currentconfidence,
                                    ) = event_locations_size_dict[
                                        int(currenttime),
                                        int(nearest_location[0]),
                                        int(nearest_location[1]),
                                        int(nearest_location[2]),
                                    ]
                                    if currentscore >= forwardscore:
                                        event_locations_size_dict.pop(
                                            (
                                                int(forwardtime),
                                                int(location[0]),
                                                int(location[1]),
                                                int(location[2]),
                                            )
                                        )

                                    if currentscore < forwardscore:
                                        event_locations_size_dict.pop(
                                            (
                                                int(currenttime),
                                                int(nearest_location[0]),
                                                int(nearest_location[1]),
                                                int(nearest_location[2]),
                                            )
                                        )
                                        
        return event_locations_size_dict                                

def headlesscall(
    key_categories: dict,
    event_threshold: float,
    nms_space: int,
    nms_time: int,
    csvdir: str,
    savedir: str,
):
    for (event_name, event_label) in key_categories.items():
        if event_label > 0:
            event_locations = []
            size_locations = []
            score_locations = []
            event_locations = []
            confidence_locations = []
            event_locations_dict = {}
            event_locations_size_dict = {}
            csvnames = list(Path(csvdir).glob("*.csv"))
            for csvname in csvnames:
                event_locations = []
                size_locations = []
                score_locations = []
                event_locations = []
                confidence_locations = []
                event_locations_dict = {}
                event_locations_size_dict = {}
                savename = Path(csvname).stem
                print(savename)
                dataset = pd.read_csv(csvname, delimiter=",")
                # Data is written as T, Y, X, Score, Size, Confidence
                T = dataset[dataset.keys()[0]][0:]
                Y = dataset[dataset.keys()[2]][0:]
                X = dataset[dataset.keys()[3]][0:]
                Score = dataset[dataset.keys()[4]][0:]
                Size = dataset[dataset.keys()[5]][0:]
                Confidence = dataset[dataset.keys()[6]][0:]
                listtime = T.tolist()
                listy = Y.tolist()
                listx = X.tolist()
                listsize = Size.tolist()

                listscore = Score.tolist()
                listconfidence = Confidence.tolist()

                for i in range(len(listtime)):

                    tcenter = int(listtime[i])
                    ycenter = float(listy[i])
                    xcenter = float(listx[i])
                    size = float(listsize[i])
                    score = float(listscore[i])
                    confidence = listconfidence[i]
                    if score > event_threshold:
                        event_locations.append(
                            [int(tcenter), int(ycenter), int(xcenter)]
                        )

                        if int(tcenter) in event_locations_dict.keys():
                            current_list = event_locations_dict[int(tcenter)]
                            current_list.append([int(ycenter), int(xcenter)])
                            event_locations_dict[int(tcenter)] = current_list
                            event_locations_size_dict[
                                (int(tcenter), int(ycenter), int(xcenter))
                            ] = [size, score]
                        else:
                            current_list = []
                            current_list.append([int(ycenter), int(xcenter)])
                            event_locations_dict[int(tcenter)] = current_list
                            event_locations_size_dict[
                                int(tcenter), int(ycenter), int(xcenter)
                            ] = [size, score]

                        size_locations.append(size)
                        score_locations.append(score)
                        confidence_locations.append(confidence)

                event_locations_size_dict = cluster_points(
                    event_locations_dict,
                    event_locations_size_dict,
                    nms_space,
                    nms_time,
                )
                event_locations_clean = []
                dict_locations = event_locations_size_dict.keys()
                tlocations = []
                zlocations = []
                ylocations = []
                xlocations = []
                scores = []
                radiuses = []
                confidences = []
                for location, sizescore in event_locations_size_dict.items():
                    tlocations.append(float(location[0]))
                    zlocations.append(0)
                    ylocations.append(float(location[1]))
                    xlocations.append(float(location[2]))
                    scores.append(float(sizescore[1]))
                    radiuses.append(float(sizescore[0]))
                    confidences.append(1)
                for location in dict_locations:
                    event_locations_clean.append(location)

                event_count = np.column_stack(
                    [
                        tlocations,
                        zlocations,
                        ylocations,
                        xlocations,
                        scores,
                        radiuses,
                        confidences,
                    ]
                )
                event_count = sorted(
                    event_count, key=lambda x: x[0], reverse=False
                )

                event_data = []
                csvname = savedir + "/" + "clean_" + savename
                if os.path.exists(csvname + ".csv"):
                    os.remove(csvname + ".csv")
                writer = csv.writer(open(csvname + ".csv", "a", newline=""))
                filesize = os.stat(csvname + ".csv").st_size

                if filesize < 1:
                    writer.writerow(
                        [
                            "T",
                            "Z",
                            "Y",
                            "X",
                            "Score",
                            "Size",
                            "Confidence",
                        ]
                    )
                for line in event_count:
                    if line not in event_data:
                        event_data.append(line)
                    writer.writerows(event_data)
                    event_data = []


def headlessvolumecall(
    key_categories: dict,
    event_threshold: float,
    nms_space: int,
    nms_time: int,
    csvdir: str,
    savedir: str,
):
    for (event_name, event_label) in key_categories.items():
        if event_label > 0:
            event_locations = []
            size_locations = []
            score_locations = []
            event_locations = []
            confidence_locations = []
            event_locations_dict = {}
            event_locations_size_dict = {}
            csvnames = list(Path(csvdir).glob("*.csv"))
            for csvname in csvnames:
                event_locations = []
                size_locations = []
                score_locations = []
                event_locations = []
                confidence_locations = []
                event_locations_dict = {}
                event_locations_size_dict = {}
                savename = Path(csvname).stem
                print(savename)
                dataset = pd.read_csv(csvname, delimiter=",")
                # Data is written as T, Y, X, Score, Size, Confidence
                T = dataset[dataset.keys()[0]][0:]
                Z = dataset[dataset.keys()[1]][0:]
                Y = dataset[dataset.keys()[2]][0:]
                X = dataset[dataset.keys()[3]][0:]
                Score = dataset[dataset.keys()[4]][0:]
                Size = dataset[dataset.keys()[5]][0:]
                Confidence = dataset[dataset.keys()[6]][0:]
                listtime = T.tolist()
                listy = Y.tolist()
                listz = Z.tolist()
                listx = X.tolist()
                listsize = Size.tolist()

                listscore = Score.tolist()
                listconfidence = Confidence.tolist()

                for i in range(len(listtime)):

                    tcenter = int(listtime[i])
                    ycenter = listy[i]
                    zcenter = listz[i]
                    xcenter = listx[i]
                    size = listsize[i]
                    score = listscore[i]
                    confidence = listconfidence[i]
                    if score > event_threshold:
                        event_locations.append(
                            [int(tcenter), int(zcenter), int(ycenter), int(xcenter)]
                        )

                        if int(tcenter) in event_locations_dict.keys():
                            current_list = event_locations_dict[int(tcenter)]
                            current_list.append([int(zcenter), int(ycenter), int(xcenter)])
                            event_locations_dict[int(tcenter)] = current_list
                            event_locations_size_dict[
                                (int(tcenter), int(zcenter), int(ycenter), int(xcenter))
                            ] = [size, score]
                        else:
                            current_list = []
                            current_list.append([int(zcenter), int(ycenter), int(xcenter)])
                            event_locations_dict[int(tcenter)] = current_list
                            event_locations_size_dict[
                                int(tcenter), int(zcenter), int(ycenter), int(xcenter)
                            ] = [size, score]

                        size_locations.append(size)
                        score_locations.append(score)
                        confidence_locations.append(confidence)

                event_locations_size_dict = cluster_points(
                    event_locations_dict,
                    event_locations_size_dict,
                    nms_space,
                    nms_time,
                )
                event_locations_clean = []
                dict_locations = event_locations_size_dict.keys()
                tlocations = []
                zlocations = []
                ylocations = []
                xlocations = []
                scores = []
                radiuses = []
                confidences = []
                for location, sizescore in event_locations_size_dict.items():
                    tlocations.append(float(location[0]))
                    zlocations.append(float(location[1]))
                    ylocations.append(float(location[2]))
                    xlocations.append(float(location[3]))
                    scores.append(float(sizescore[1]))
                    radiuses.append(float(sizescore[0]))
                    confidences.append(1)
                for location in dict_locations:
                    event_locations_clean.append(location)

                event_count = np.column_stack(
                    [
                        tlocations,
                        zlocations,
                        ylocations,
                        xlocations,
                        scores,
                        radiuses,
                        confidences,
                    ]
                )
                event_count = sorted(
                    event_count, key=lambda x: x[0], reverse=False
                )

                event_data = []
                csvname = savedir + "/" + "clean_" + savename
                if os.path.exists(csvname + ".csv"):
                    os.remove(csvname + ".csv")
                writer = csv.writer(open(csvname + ".csv", "a", newline=""))
                filesize = os.stat(csvname + ".csv").st_size

                if filesize < 1:
                    writer.writerow(
                        [
                            "T",
                            "Z",
                            "Y",
                            "X",
                            "Score",
                            "Size",
                            "Confidence",
                        ]
                    )
                for line in event_count:
                    if line not in event_data:
                        event_data.append(line)
                    writer.writerows(event_data)
                    event_data = []


from napari.viewer import Viewer
from qtpy.QtWidgets import (QWidget, QVBoxLayout, QLabel, QFileDialog)
from .OneatVisWidget import OneatVisWidget
from ..OneatSimpleVisualization import OneatSimpleVisualization

class OneatVisualizeWidget(QWidget):
    
    def __init__(self,
                 viewer : Viewer,
                 parent = None):
        
        super().__init__(parent = parent)
    
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)
        
        self._layout.addWidget(QLabel("Visualize Detections", parent = self))
        self.viswidget = OneatVisWidget(parent = self)
        self._layout.addWidget(self.viswidget)
        self.event_threshold = float(self.viswidget.label.text())
        self.start_prob = self.viswidget.startprobspinbox.value()
        self.viswidget.startprobspinbox.valueChanged.connect(self._update_startprob_callback)
        self.viswidget.scoreslider.valueChanged.connect(self._update_slider_callback)
        
        self.simplevisualization = OneatSimpleVisualization(viewer,  self.viswidget.ax, self.viswidget.figure )
        
       
        
        self.viswidget.detectionidbox.clicked.connect(
            lambda detectioid = self.viswidget.detectionidbox: self._capture_detections_callback()
        )
        self.viswidget.recomputebutton.clicked.connect(
            lambda eventid=self.viswidget.recomputebutton: self._recompute_callback()
        )
    
    def _update_startprob_callback(self, event):
        self.start_prob = self.viswidget.startprobspinbox.value()    
        self.event_threshold = float(self.viswidget.label.text())

    def _update_slider_callback(self, value):

        real_value = float(
            self.start_prob + (1.0 - self.start_prob) / 5000 * float(value)
        )
        self.viswidget.label.setText(str(real_value))
        self.event_threshold = float(real_value)    
        
   
    def _recompute_callback(self):  
        
          if self.csvname is not None:
                self.event_threshold = float(self.viswidget.label.text())
                self.simplevisualization.show_csv(self.csvname,self.event_threshold)       
           
    def _capture_detections_callback(self):  
        
          self.csvname = QFileDialog.getOpenFileName(self, "Open oneat detections file")
          if self.csvname is not None:
                self.csvname = self.csvname[0]
                self.event_threshold = float(self.viswidget.label.text())
                self.simplevisualization.show_csv(self.csvname,self.event_threshold)
                        


#!/usr/bin/env python3
"""
Created on Sat May 23 15:13:01 2020

@author: aimachine
"""

import os

# from IPython.display import clear_output
from pathlib import Path

import numpy as np
import tensorflow as tf
from keras import callbacks
from tensorflow.keras.models import load_model
from tifffile import imread
from tqdm import tqdm
import datetime
from oneat.NEATModels import nets
from oneat.NEATModels.loss import class_yolo_loss, static_yolo_loss
from oneat.NEATModels.nets import Concat
from oneat.NEATUtils import plotters, utils
from oneat.NEATUtils.utils import (
    goodboxes,
    load_json,
    normalizeFloatZeroOne,
    save_static_csv,
    save_volume,
    yoloprediction,
)

Boxname = "ImageIDBox"
CellTypeBoxname = "CellIDBox"


class NEATResNet:
    """
    Parameters
    ----------

    NpzDirectory : Specify the location of npz file containing the training data with movies and labels

    TrainModelName : Specify the name of the npz file containing training data and labels

    ValidationModelName :  Specify the name of the npz file containing validation data and labels

    categories : Number of action classes

    Categories_Name : List of class names and labels

    model_dir : Directory location where trained model weights are to be read or written from


    model_keras : The model as it appears as a Keras function

    model_weights : If re-training model_weights = model_dir  else None as default

    lstm_hidden_units : Number of hidden uniots for LSTm layer, 64 by default

    epochs :  Number of training epochs, 55 by default

    batch_size : batch_size to be used for training, 20 by default



    """

    def __init__(
        self,
        staticconfig,
        model_dir,
        catconfig,
        cordconfig,
        class_only=False,
        train_lstm=False,
    ):

        self.staticconfig = staticconfig
        self.catconfig = catconfig
        self.cordconfig = cordconfig
        self.model_dir = model_dir
        self.class_only = class_only
        self.train_lstm = train_lstm
        self.key_cord = self.cordconfig
        self.categories = len(self.catconfig)
        self.key_categories = self.catconfig
        if self.staticconfig is not None:
            self.npz_directory = staticconfig.npz_directory
            self.npz_name = staticconfig.npz_name
            self.npz_val_name = staticconfig.npz_val_name
            self.key_categories = staticconfig.key_categories
            self.box_vector = staticconfig.box_vector
            self.show = staticconfig.show
            self.key_cord = staticconfig.key_cord
            self.categories = len(staticconfig.key_categories)
            self.stage_number = staticconfig.stage_number
            self.last_conv_factor = 2 ** (self.stage_number - 1)
            self.depth = staticconfig.depth
            self.start_kernel = staticconfig.start_kernel
            self.mid_kernel = staticconfig.mid_kernel
            self.learning_rate = staticconfig.learning_rate
            self.epochs = staticconfig.epochs
            self.startfilter = staticconfig.startfilter
            self.batch_size = staticconfig.batch_size
            self.multievent = staticconfig.multievent
            self.imagex = staticconfig.imagex
            self.imagey = staticconfig.imagey
            self.nboxes = staticconfig.nboxes
            self.gridx = staticconfig.gridx
            self.gridy = staticconfig.gridy
            self.yolo_v0 = staticconfig.yolo_v0
            self.stride = staticconfig.stride

        if self.staticconfig is None:

            self.staticconfig = load_json(
                os.path.join(self.model_dir, "parameters.json")
            )

            self.npz_directory = self.staticconfig["npz_directory"]
            self.npz_name = self.staticconfig["npz_name"]
            self.npz_val_name = self.staticconfig["npz_val_name"]
            self.box_vector = self.staticconfig["box_vector"]
            self.show = self.staticconfig["show"]
            self.depth = self.staticconfig["depth"]
            self.start_kernel = self.staticconfig["start_kernel"]
            self.mid_kernel = self.staticconfig["mid_kernel"]
            self.learning_rate = self.staticconfig["learning_rate"]
            self.epochs = self.staticconfig["epochs"]
            self.startfilter = self.staticconfig["startfilter"]
            self.batch_size = self.staticconfig["batch_size"]
            self.multievent = self.staticconfig["multievent"]
            self.imagex = self.staticconfig["imagex"]
            self.imagey = self.staticconfig["imagey"]
            self.nboxes = self.staticconfig["nboxes"]
            self.gridx = self.staticconfig["gridx"]
            self.gridy = self.staticconfig["gridy"]
            self.yolo_v0 = self.staticconfig["yolo_v0"]
            self.stride = self.staticconfig["stride"]

            self.stage_number = self.staticconfig["stage_number"]
            self.last_conv_factor = 2 ** (self.stage_number - 1)

        self.X = None
        self.Y = None
        self.axes = None
        self.X_val = None
        self.Y_val = None
        self.Trainingmodel = None
        self.Xoriginal = None
        self.Xoriginal_val = None
        if not self.class_only:

            if not self.train_lstm:
                self.model_keras = nets.resnet_v2

            if self.train_lstm:
                self.model_keras = nets.resnet_lstm_v2

            if self.multievent:
                self.last_activation = "sigmoid"
                self.entropy = "binary"

            if not self.multievent:
                self.last_activation = "softmax"
                self.entropy = "notbinary"

            self.yolo_loss = static_yolo_loss(
                self.categories,
                self.gridx,
                self.gridy,
                self.nboxes,
                self.box_vector,
                self.entropy,
                self.yolo_v0,
            )

        if self.class_only:

            if not self.train_lstm:
                self.model_keras = nets.resnet_v2_class
            if self.train_lstm:
                self.model_keras = nets.resnet_lstm_v2_class

            if self.multievent:
                self.last_activation = "sigmoid"
                self.entropy = "binary"

            if not self.multievent:
                self.last_activation = "softmax"
                self.entropy = "notbinary"

            self.yolo_loss = class_yolo_loss(self.categories, self.entropy)

    def loadData(self, sum_channels=False):

        (X, Y), axes = utils.load_full_training_data(
            self.npz_directory, self.npz_name, verbose=True
        )

        (X_val, Y_val), axes = utils.load_full_training_data(
            self.npz_directory, self.npz_val_name, verbose=True
        )

        self.Xoriginal = X
        self.Xoriginal_val = X_val

        self.X = X
        if sum_channels:
            self.X = np.sum(X, -1)
            self.X = np.expand_dims(self.X, -1)
        self.Y = Y[:, :, 0]
        self.X_val = X_val
        if sum_channels:
            self.X_val = np.sum(X_val, -1)
            self.X_val = np.expand_dims(self.X_val, -1)
        self.Y_val = Y_val[:, :, 0]
        self.axes = axes
        self.Y = self.Y.reshape((self.Y.shape[0], 1, 1, self.Y.shape[1]))
        self.Y_val = self.Y_val.reshape(
            (self.Y_val.shape[0], 1, 1, self.Y_val.shape[1])
        )
        print(self.X.shape, self.Y.shape)

    def TrainModel(self):

        input_shape = (self.X.shape[1], self.X.shape[2], self.X.shape[3])

        Path(self.model_dir).mkdir(exist_ok=True)

        Y_main = self.Y[:, :, :, 0 : self.categories - 1]

        y_integers = np.argmax(Y_main, axis=-1)
        y_integers = y_integers[:, 0, 0]

        dummyY = np.zeros(
            [
                self.Y.shape[0],
                self.Y.shape[1],
                self.Y.shape[2],
                self.categories + self.nboxes * self.box_vector,
            ]
        )
        dummyY[:, :, :, : self.Y.shape[3]] = self.Y

        dummyY_val = np.zeros(
            [
                self.Y_val.shape[0],
                self.Y_val.shape[1],
                self.Y_val.shape[2],
                self.categories + self.nboxes * self.box_vector,
            ]
        )
        dummyY_val[:, :, :, : self.Y_val.shape[3]] = self.Y_val
        for b in range(1, self.nboxes):
            dummyY[
                :,
                :,
                :,
                self.categories
                + b * self.box_vector : self.categories
                + (b + 1) * self.box_vector,
            ] = self.Y[
                :, :, :, self.categories : self.categories + self.box_vector
            ]
            dummyY_val[
                :,
                :,
                :,
                self.categories
                + b * self.box_vector : self.categories
                + (b + 1) * self.box_vector,
            ] = self.Y_val[
                :, :, :, self.categories : self.categories + self.box_vector
            ]

        self.Y = dummyY
        self.Y_val = dummyY_val

        print(self.Y.shape, self.nboxes)

        self.Trainingmodel = self.model_keras(
            input_shape,
            self.categories,
            box_vector=self.box_vector,
            yolo_loss = self.yolo_loss, 
            nboxes=self.nboxes,
            stage_number=self.stage_number,
            depth=self.depth,
            start_kernel=self.start_kernel,
            mid_kernel=self.mid_kernel,
            startfilter=self.startfilter,
            last_activation=self.last_activation,
            input_model=self.model_dir,
        )

        sgd = tf.keras.optimizers.Adam(learning_rate=self.learning_rate)
        self.Trainingmodel.compile(
            optimizer=sgd, loss=self.yolo_loss, metrics=["accuracy"]
        )
        self.Trainingmodel.summary()
        
        # Keras callbacks
        lrate = callbacks.ReduceLROnPlateau(
            monitor="loss", factor=0.1, patience=4, verbose=1
        )
        hrate = callbacks.History()
        srate = callbacks.ModelCheckpoint(
            self.model_dir,
            monitor="loss",
            verbose=1,
            save_best_only=False,
            save_weights_only=False,
            mode="auto",
            period=1,
        )
        prate = plotters.PlotStaticHistory(
            self.Trainingmodel,
            self.X_val,
            self.Y_val,
            self.key_categories,
            self.key_cord,
            self.gridx,
            self.gridy,
            plot=self.show,
            nboxes=self.nboxes,
            class_only=self.class_only,
        )
        log_dir = "logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
        tensorboard_callback = callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1)
        # Train the model and save as a h5 file
        self.Trainingmodel.fit(
            self.X,
            self.Y,
            batch_size=self.batch_size,
            epochs=self.epochs,
            validation_data=(self.X_val, self.Y_val),
            shuffle=True,
            callbacks=[lrate, hrate, srate, prate, tensorboard_callback],
        )

        self.Trainingmodel.save(self.model_dir)

    def predict(
        self,
        image: np.ndarray,
        savedir: str = None,
        event_threshold: float = 0.5,
        event_confidence: float = 0.5,
        n_tiles: tuple = (1, 1),
        overlap_percent: float = 0.8,
        iou_threshold: float = 0.01,
        RGB: bool = False,
        fidelity: int = 1,
        normalize: bool = True,
        activations=False,
    ):

        self.image = imread(image)

        self.savedir = savedir
        self.n_tiles = n_tiles
        self.RGB = RGB
        self.fidelity = fidelity
        self.overlap_percent = overlap_percent
        self.iou_threshold = iou_threshold
        self.event_threshold = event_threshold
        self.event_confidence = event_confidence
        self.originalimage = self.image
        self.activations = activations

        self.normalize = normalize

        self.model = self._build()

        eventboxes = []
        classedboxes = {}
        self.iou_classedboxes = {}
        self.all_iou_classedboxes = {}
        if self.normalize:
            self.image = normalizeFloatZeroOne(self.image, 1, 99.8)

        if not RGB:
            for inputtime in tqdm(range(0, self.image.shape[0])):

                smallimage = self.image[inputtime, :]

                # Break image into tiles if neccessary
                predictions, allx, ally = self.predict_main(smallimage)

                # Iterate over tiles
                for p in range(0, len(predictions)):

                    sum_time_prediction = predictions[p]

                    if sum_time_prediction is not None:
                        # For each tile the prediction vector has shape N H W Categories + Trainng Vector labels
                        for i in range(0, sum_time_prediction.shape[0]):
                            time_prediction = sum_time_prediction[i]
                            boxprediction = yoloprediction(
                                ally[p],
                                allx[p],
                                time_prediction,
                                self.stride,
                                inputtime,
                                self.staticconfig,
                                self.key_categories,
                                self.key_cord,
                                self.nboxes,
                                "detection",
                                "static",
                            )

                            if boxprediction is not None:
                                eventboxes = eventboxes + boxprediction

                for (event_name, event_label) in self.key_categories.items():

                    if event_label > 0:
                        current_event_box = []
                        for box in eventboxes:

                            event_prob = box[event_name]
                            event_confidence = box["confidence"]
                            if (
                                event_prob >= self.event_threshold[event_label]
                                and event_confidence >= self.event_confidence
                            ):
                                current_event_box.append(box)
                        classedboxes[event_name] = [current_event_box]

                self.classedboxes = classedboxes
                self.eventboxes = eventboxes

                self.nms()
                if self.savedir is not None:
                    self.to_csv()
                if self.activations:
                    self.to_activations()
                eventboxes = []
                classedboxes = {}

        if RGB:

            smallimage = self.image[:, :, 0]
            if self.normalize:
                smallimage = normalizeFloatZeroOne(smallimage, 1, 99.8)
            # Break image into tiles if neccessary
            predictions, allx, ally = self.predict_main(smallimage)

            # Iterate over tiles
            for p in range(0, len(predictions)):

                sum_time_prediction = predictions[p]

                if sum_time_prediction is not None:
                    # For each tile the prediction vector has shape N H W Categories + Trainng Vector labels
                    for i in range(0, sum_time_prediction.shape[0]):
                        time_prediction = sum_time_prediction[i]

                        boxprediction = yoloprediction(
                            ally[p],
                            allx[p],
                            time_prediction,
                            self.stride,
                            0,
                            self.staticconfig,
                            self.key_categories,
                            self.key_cord,
                            self.nboxes,
                            "detection",
                            "static",
                        )

                        if boxprediction is not None:
                            eventboxes = eventboxes + boxprediction

            for (event_name, event_label) in self.key_categories.items():

                if event_label > 0:
                    current_event_box = []
                    for box in eventboxes:

                        event_prob = box[event_name]
                        event_confidence = box["confidence"]
                        if (
                            event_prob >= self.event_threshold
                            and event_confidence >= self.event_confidence
                        ):
                            current_event_box.append(box)
                    classedboxes[event_name] = [current_event_box]

            self.classedboxes = classedboxes
            self.eventboxes = eventboxes
            # self.iou_classedboxes = classedboxes
            self.nms()
            if self.savedir is not None:
                self.to_csv()
            if self.activations:
                self.to_activations()
            eventboxes = []
            classedboxes = {}

    def _build(self):

        Model = load_model(
            self.model_dir,
            custom_objects={"loss": self.yolo_loss, "Concat": Concat},
        )
        return Model

    def nms(self):

        best_iou_classedboxes = {}
        self.iou_classedboxes = {}
        for (event_name, event_label) in self.key_categories.items():
            if event_label > 0:
                # Get all events
                sorted_event_box = self.classedboxes[event_name][0]
                scores = [
                    sorted_event_box[i][event_name]
                    for i in range(len(sorted_event_box))
                ]
                best_sorted_event_box = goodboxes(
                    sorted_event_box,
                    scores,
                    self.iou_threshold,
                    self.event_threshold[event_label],
                    self.imagex,
                    self.imagey,
                    fidelity=self.fidelity,
                )

                best_iou_classedboxes[event_name] = [best_sorted_event_box]

        self.iou_classedboxes = best_iou_classedboxes

    def to_csv(self):

        save_static_csv(
            self.key_categories, self.iou_classedboxes, self.savedir
        )

    def to_activations(self):

        self.all_iou_classedboxes = save_volume(
            self.key_categories,
            self.iou_classedboxes,
            self.all_iou_classedboxes,
        )

    def overlaptiles(self, sliceregion):

        if self.n_tiles == (1, 1):

            patchshape = (sliceregion.shape[0], sliceregion.shape[1])

            patch = []
            rowout = []
            column = []

            patch.append(sliceregion)
            rowout.append(0)
            column.append(0)

        else:

            patchx = sliceregion.shape[1] // self.n_tiles[0]
            patchy = sliceregion.shape[0] // self.n_tiles[1]

            if patchx > self.imagex and patchy > self.imagey:
                if self.overlap_percent > 1 or self.overlap_percent < 0:
                    self.overlap_percent = 0.8

                jumpx = int(self.overlap_percent * patchx)
                jumpy = int(self.overlap_percent * patchy)

                patchshape = (patchy, patchx)
                rowstart = 0
                colstart = 0
                pairs = []
                # row is y, col is x

                while rowstart < sliceregion.shape[0] - patchy:
                    colstart = 0
                    while colstart < sliceregion.shape[1] - patchx:
                        # Start iterating over the tile with jumps = stride of the fully convolutional network.
                        pairs.append([rowstart, colstart])
                        colstart += jumpx
                    rowstart += jumpy

                    # Include the last patch
                rowstart = sliceregion.shape[0] - patchy
                colstart = 0
                while colstart < sliceregion.shape[1]:
                    pairs.append([rowstart, colstart])
                    colstart += jumpx
                rowstart = 0
                colstart = sliceregion.shape[1] - patchx
                while rowstart < sliceregion.shape[0]:
                    pairs.append([rowstart, colstart])
                    rowstart += jumpy

                if (
                    sliceregion.shape[0] >= self.imagey
                    and sliceregion.shape[1] >= self.imagex
                ):

                    patch = []
                    rowout = []
                    column = []
                    for pair in pairs:
                        smallpatch, smallrowout, smallcolumn = chunk_list(
                            sliceregion, patchshape, pair
                        )
                        patch.append(smallpatch)
                        rowout.append(smallrowout)
                        column.append(smallcolumn)

            else:

                patch = []
                rowout = []
                column = []
                patch.append(sliceregion)
                rowout.append(0)
                column.append(0)

        self.patch = patch
        self.sy = rowout
        self.sx = column

    def predict_main(self, sliceregion):
        try:
            self.overlaptiles(sliceregion)
            predictions = []
            allx = []
            ally = []
            if len(self.patch) > 0:
                for i in range(0, len(self.patch)):
                    sum_time_prediction = self.make_patches(self.patch[i])

                    predictions.append(sum_time_prediction)
                    allx.append(self.sx[i])
                    ally.append(self.sy[i])

            else:

                sum_time_prediction = self.make_patches(self.patch)
                predictions.append(sum_time_prediction)
                allx.append(self.sx)
                ally.append(self.sy)

        except tf.errors.ResourceExhaustedError:

            print("Out of memory, increasing overlapping tiles for prediction")

            self.list_n_tiles = list(self.n_tiles)
            self.list_n_tiles[0] = self.n_tiles[0] + 1
            self.list_n_tiles[1] = self.n_tiles[1] + 1
            self.n_tiles = tuple(self.list_n_tiles)
            self.predict_main(sliceregion)

        return predictions, allx, ally

    def make_patches(self, sliceregion):

        predict_im = np.expand_dims(sliceregion, 0)

        prediction_vector = self.model.predict(
            np.expand_dims(predict_im, -1), verbose=0
        )

        return prediction_vector


def chunk_list(image, patchshape, pair):
    rowstart = pair[0]
    colstart = pair[1]

    endrow = rowstart + patchshape[0]
    endcol = colstart + patchshape[1]

    if endrow > image.shape[0]:
        endrow = image.shape[0]
    if endcol > image.shape[1]:
        endcol = image.shape[1]

    region = (slice(rowstart, endrow), slice(colstart, endcol))

    # The actual pixels in that region.
    patch = image[region]
    # Always normalize patch that goes into the netowrk for getting a prediction score

    return patch, rowstart, colstart
